/*
Don's Up/Down Timer
JavaScript count up/down timer, version 0.99
June 2011 - January 2013
Donald Byrd
Woodrow Wilson Indiana Teaching Fellow
Indiana University School of Informatics / IUPUI Mathematics Education
donbyrd@indiana.edu


I call this little program a "count up/down" timer (as opposed to a "countdown" or
"count-down/up" timer, etc.) because it's designed for use in situations where you want
to measure an unpredictable interval of time, then set a timer to go off when the same
interval elapses again. One such situation is in a K-12 classroom, where students are
misbehaving and and, to discourage them, the teacher wants to make them do something
they won't like (e.g., stay after class) for the same amount of time. I wanted a program
to do this, and preferably written in JavaScript, so it could run in almost any Web
browser with no installation. Surely programs already existed that can do it, probably
even programs in JavaScript; but I couldn't find one, so here's mine!

I've tested this program on OS X with Firefox and Safari. I don't expect problems with
other platforms/web browsers, but it really should be tested more.

Don's Up/Down Timer owes a lot, including code, to the JavaScript Countdown/Count-up
Timer/Clock/Ticker for Web Pages by Robert Hashemian (http://www.hashemian.com/).
Hashemian says of his work: "You can use this code in any manner so long as the author's
name, Web address and this disclaimer is kept intact."

Options are:
* TimeFormat allows you to format the countdown/count-up display to your liking. For
	example, instead of the default English, you can use terms from another language, or
	make adjustments like omitting the seconds segment or just using colons between time
	units.
* TimeColor sets the color of characters in the time displayed.

* FinishMessage: message to display (or "" for no message) when the countdown reaches
	zero. (Obviously never displayed when counting up.)
* FinishColor sets the color of characters in FinishMessage.
* FinishSound: play sound when countdown reaches zero.

* AutoCountdownTime is a string giving the time of day at which counting down should begin
	automatically.
* AutoCountdownColor sets the color of characters in the time display once auto countdown begins.

* LeadingZero: true = display leading zeros; false = don't display them.
* BoldDigits: true = display digits in bold (other characters displayed aren't affected).
* CanSkip: true = time displayed can skip seconds to maintain accuracy; see comments on
	function updateTimeDisp() below.

Any or all can be omitted; the default behavior is reasonable (in my opinion :-) ).

Usage Sample:

<script language="JavaScript">
TimeFormat = "%%M min. %%S sec.";
TimeColor = "black";
LeadingZero = true;
BoldDigits = true;
CanSkip = true;
FinishMessage = "<b>Time is up.</b>";
FinishColor = "darkred";
</script>
<script language="JavaScript" src="DonsUpDownTimer.js"></script>

Note that you can control lots of other things by making obvious changes, and more with
a little knowledge of HTML; see below the heading "Run the Up/Down Timer". 

TO DO IDEAS: check the code with JSLint and/or the new "use strict" pramga and clean up;
improve error checking, e.g., for "Set" number; test more with other browsers, esp.
Internet Explorer; add BackColor option (a la Hashemian); add time scale for countdown;
etc. Suggestions are very welcome; email me!  --DAB

*/

// ------------------------------------- Up/Down Timer Code -------------------------------
// JavaScript code for the Up/Down Timer. Changing any of this could lead to unpleasant
// surprises unless you know what you're doing!

var theIntervalID = 0;
var numSeconds = 0;
var countStep = 0;
var prevSetNumStr = "0";
var doAutoCountdown = false;	// Has automatically-started countdown started?
var debug = false;
var nFinishSoundReps = 1;		// No. of times to repeat FinishSound -- but see below!

// Stop updating the time remaining
function pauseIntervalTimer()
{
    countStep = 0;
}

function sayIfAutoCountdownCancelled()
{
	if (doAutoCountdown)
		document.getElementById("autoStartWhen").innerHTML = autoStartStr + "  (CANCELLED)";
}

/* Functions to handle clicks on the buttons */
function handleGoUpClick()
{
    pauseIntervalTimer();
    // Set the time increment to go up, and start the timer
	countStep = 1;
    //theIntervalID = setInterval("updateTime()", 1000);
    datimeNow = new Date();
	sayIfAutoCountdownCancelled();
    document.tm.pause.disabled = false;
}

function handleGoDownClick()
{
    pauseIntervalTimer();
    // Set the time increment to go down, and start the timer
	countStep = -1;
    //theIntervalID = setInterval("updateTime()", 1000);
    datimeNow = new Date();
	sayIfAutoCountdownCancelled();
    document.tm.pause.disabled = false;
}

function handlePauseClick()
{
    document.tm.goUp.disabled = false;
    document.tm.goDown.disabled = false;
    this.disabled = true;
    pauseIntervalTimer();
}

function handleSetClick()
{
	var nonDigitsRE = /[^\-0-9]/;
    var newNumStr;
    document.tm.goUp.disabled = false;
    document.tm.goDown.disabled = false;

    // Get new time counter value from user; check it for legality; and, if it's okay, set
    //   the timer and display to the new value. But first, stop the timer: even if they
    //   enter an illegal value, the user would probably want it stopped.
    pauseIntervalTimer();
    newNumStr = prompt("Enter new time display value (in seconds):", prevSetNumStr);
    if (!newNumStr) {
	  alert("ERROR: Nothing was entered for the new time display value.");
    }
    // We have a string. If it's a valid non-negative integer, set the counter to it; else
    //  give an error message and ignore it.
    if (newNumStr) {
      nondigitPos = newNumStr.search(nonDigitsRE);
      if (nondigitPos>=0) {
      	alert("ERROR: New time display value '" + newNumStr + "' is illegal: it must be a non-negative integer.");
      	return;
      }
	  newNumSeconds = parseInt(newNumStr, 10);
	  if (isNaN(newNumSeconds)) alert("ERROR: New time display value '" + newNumStr + "' is illegal: it must be an integer.");
	  else if (newNumSeconds<0) alert("ERROR: New time display value '" + newNumStr + "' is illegal: it cannot be negative.");
	  else {
	    prevSetNumStr = newNumStr;
	    numSeconds = newNumSeconds;
	    showTime(numSeconds);
	    datimeNow = new Date();
	    sayIfAutoCountdownCancelled();
	  }
	}
}

/* Return secs/sDiv modulo sMod as a string */
function calcAge(secs, sDiv, sMod) {
  var s = ((Math.floor(secs/sDiv))%sMod).toString();
  if (LeadingZero && s.length < 2) s = "0" + s;
  if (BoldDigits) return "<b>" + s + "</b>";
  else return s;
}

function showTime(nSeconds) {
  var dispStr, theColor;

  dispStr = TimeFormat.replace(/%%D/g, calcAge(nSeconds,86400,100000));
  dispStr = dispStr.replace(/%%H/g, calcAge(nSeconds,3600,24));
  dispStr = dispStr.replace(/%%M/g, calcAge(nSeconds,60,60));
  dispStr = dispStr.replace(/%%S/g, calcAge(nSeconds,1,60));

  theColor = dispStr.fontcolor(TimeColor);
  if (doAutoCountdown)
    theColor = dispStr.fontcolor(AutoCountdownColor);

  document.getElementById("timeMessage").innerHTML = theColor;
}

function formatTime(aDate) {
	var hours = aDate.getHours();
	var minutes = aDate.getMinutes();
	if (minutes<10) minutes = "0" + minutes;
	suffix = " AM";
	if (hours>12) {
		hours -= 12;
		suffix = " PM";
	}
	timeStr = hours + ":" + minutes + suffix;
	return timeStr;
}

/* Do everything necessary to initialize the program: mostly setting unused options
     to default values. */
function initTimer()
{
	if (typeof(TimeFormat)==="undefined")
//	  TimeFormat = "%%D days %%H hours %%M minutes %%S seconds";
	  TimeFormat = "%%M min. %%S sec.";
	if (typeof(TimeColor)==="undefined")
	  TimeColor = "black";

	if (typeof(FinishMessage)==="undefined")
	  FinishMessage = "Time is up!";
	if (typeof(FinishColor)==="undefined")
	  FinishColor = "darkred";
	if (typeof(FinishSound)==="undefined")
	  FinishSound = false;

	if (typeof(LeadingZero)==="undefined")
	  LeadingZero = true;
	if (typeof(BoldDigits)==="undefined")
	  BoldDigits = true;
	if (typeof(CanSkip)==="undefined")
	  CanSkip = true;

	if (typeof(AutoCountdownTime)=="undefined")
	  dAutoStart = new Date("2999/12/31");
	else {
	  // Get current date/time; set auto-countdown date to today. Note that if that
	  //   time has already passed, the program will decide as soon it's launched that
	  //   the countdown is over and display FinishMessage, while the user might have
	  //   wanted it to assume the given time tomorrow, not today. Oh well.
	  datimeNow = new Date();
	  //dAutoStart = new Date(year, month, day, hours, minutes, seconds, milliseconds);
	  dAutoStart = datimeNow;
	  delimPos = AutoCountdownTime.indexOf(':');
	  hours = AutoCountdownTime.substring(0, delimPos);
	  minutes = AutoCountdownTime.substring(delimPos+1);
	  dAutoStart.setHours(hours);
	  dAutoStart.setMinutes(minutes);
	  dAutoStart.setSeconds(0);
	  

	  if (debug)
	    document.getElementById("autoStartWhen").innerHTML = dAutoStart;
	  else {
	    autoStartStr = formatTime(dAutoStart);
	    document.getElementById("autoStartWhen").innerHTML = autoStartStr;
	  }
	  dDiff = new Date(datimeNow-dAutoStart);
	  if (dDiff>0.0) alert("ERROR: auto-countdown time has already passed.");
	}
	if (typeof(AutoCountdownColor)==="undefined")
	  AutoCountdownColor = "blue";

    // I thought the default state of buttons was always enabled, but I've seen one situation
    // (in Firefox, after reloading) where they're sometimes disabled, so play it safe.
    document.tm.goUp.disabled = false;
    document.tm.goDown.disabled = false;
    document.tm.pause.disabled = false;
    document.tm.Set.disabled = false;
}

/* If the automatic countdown time has passed and we're not already counting down (i.e.,
   we're counting up or doing nothing), start counting down now. */
function mayAutoCountDown()
{
    if (!doAutoCountdown) {
		datimeNow = new Date();
		dDiff = new Date(datimeNow-dAutoStart);
		gSecs = Math.floor(dDiff.valueOf()/1000);			// ??SHOULD BE A BETTER WAY TO DO THIS
		if (gSecs>0.0 && !doAutoCountdown) {
			//alert("TIME TO START COUNTDOWN!!!");
			handleGoDownClick();
			doAutoCountdown = true;
			document.getElementById("autoStartWhen").innerHTML = autoStartStr + "  (STARTED)";
		}
    }
}

function TimeIsUp() {
  var theColor;
  theColor = FinishMessage.fontcolor(FinishColor);
  document.getElementById("timeMessage").innerHTML = theColor;
  
  // If we're playing a sound, repeat the desired number of times. Caveat: Note that
  // this code doesn't wait until a previous playing is done before starting it again; it
  // just starts playing it again each time TimeIsUp() is called, which is normally once
  // per second. This is no problem in simple tests, but it could be with a different
  // sound, browser, etc.
  if (FinishSound && nFinishSoundReps>0) {
    var snd = new Audio("FinishSound.wav");
    snd.loop = false;
    snd.play();
    // As of early 2013, apparently, some browsers can play Ogg Vorbis, some can play .wav,
    // some can play neither or both! Obviously just trying to play .wav isn't ideal.
    // Since the JS constructor doesn't support fallback <source> elements, we could use,
    // e.g., (new Audio()).canPlayType("audio/ogg; codecs=vorbis") to test whether the browser
    // supports Ogg Vorbis.
    nFinishSoundReps = nFinishSoundReps-1;
  }
  document.tm.goUp.disabled = true;
  document.tm.goDown.disabled = true;
}

/* Update the time-remaining display. This should be called every second. If time remaining
   is zero or below, we're done, so disable counting either up or down. Otherwise, if
   _CanSkip_, we update it to reflect the total time that's elapsed since we started timing;
   if not, we simply assume it's been 1 sec. since the last update. While this is by no
   means guaranteed in any JavaScript environment I know of, it's _usually_ pretty close.
   (An old version of Firefox on an old G5 I use a lot is a dramatic exception, probably due
   to a memory leak; after it's run for awhile, it often pauses for as long as 10-15 sec.! 
   Oh well.) */
function updateTimeDisp()
{
  dPrev = datimeNow;
  // Update time count in the desired +/- direction by the actual no. of seconds
  if (CanSkip) {
    datimeNow = new Date();
    elapsed = (datimeNow-dPrev)/1000;
    numSeconds += countStep*elapsed;
  } else
      numSeconds += countStep;
  if (numSeconds<=0) {
    TimeIsUp();
  } else
    showTime(numSeconds);
  // Make old time disappear a bit before new time appears, so it flashes (optional)
  //setTimeout('document.getElementById("timeMessage").innerHTML = ""', 900);
  
}

/* Update things for current time: consider updating the time-remaining display and starting
	autoCountdown. This should be called (at least?) every second. */

function updateTime()
{
  if (countStep!=0) updateTimeDisp();
  
  // Consider starting automatic countdown.
  mayAutoCountDown();
}

// The top-level routine: get everything going.
function startTimer()
{
	initTimer();
    theIntervalID = setInterval("updateTime()", 1000);
	showTime(numSeconds);
}
